"""
Usage:
    zot contigs [-l LEN] <input>...

Perform a simple de Bruijn graph assemly of the input k-mers.  Only
non-branching paths in the de Bruijn graph are reported: no attempt
is made to resolve branching at all.

Options:
    -l LEN          minimum length of contig to report (default 2K)
"""

from pykmer.basics import rc, render
from pykmer.container import container
from pykmer.container.std import readKmers
from pykmer.sparse import sparse
from pykmer.bitvec import bitvec

import docopt
import array
import sys

def succ(K, xs, x):
    m = (1 << (2*K)) - 1
    y0 = (x << 2) & m
    y1 = y0 + 4
    (r0, r1) = xs.rank2(y0, y1)
    return range(r0, r1)

def main(argv):
    opts = docopt.docopt(__doc__, argv)

    L0 = None
    if opts['-l']:
        L0 = int(opts['-l'])

    for inp in opts['<input>']:
        with container(inp, 'r') as z:
            K = z.meta['K']
            L = L0
            if L is None:
                L = 2*K

            xs = array.array('L', readKmers(z))
            S = sparse(2*K, xs)

            seen = bitvec(S.count())
            for i in xrange(S.count()):
                if seen[i]:
                    continue

                x = S.select(i)
                xb = rc(K, x)
                xp = succ(K, S, xb)
                if xp == 1:
                    # x isn't the start of a contig
                    continue

                pth = [x]
                seen[i] = 1
                xn = succ(K, S, x)
                while len(xn) == 1:
                    if seen[xn[0]] == 1:
                        break
                    x = S.select(xn[0])
                    pth.append(x)
                    seen[xn[0]] = 1
                    xb = rc(K, x)
                    j = S.rank(xb)
                    seen[j] = 1
                    xn = succ(K, S, x)

                if len(pth)+K-1 < L:
                    continue

                s = [render(K, pth[0])]
                for j in xrange(1, len(pth)):
                    s.append("ACGT"[pth[j]&3])

                print '>contig_%d\n%s' % (i, ''.join(s))

if __name__ == '__main__':
    main(sys.argv[1:])



"""
Usage:
    zot merge <output> <input>...
"""

import os
import sys

import docopt

from pykmer.basics import render
from pykmer.container import container
from pykmer.container.std import readKmersAndCounts, writeKmersAndCounts
from pykmer.file import tmpfile

def pairs(xs):
    i = 0
    while i + 1 < len(xs):
        yield (xs[i], xs[i + 1])
        i += 2
    if i < len(xs):
        yield (xs[i], )

def merge(xs, ys):
    moreXs = True
    moreYs = True

    try:
        x = xs.next()
        xk = x[0]
    except StopIteration:
        moreXs = False
    try:
        y = ys.next()
        yk = y[0]
    except StopIteration:
        moreYs = False

    while moreXs and moreYs:
        if xk < yk:
            yield x
            try:
                x = xs.next()
                xk = x[0]
            except StopIteration:
                moreXs = False
            continue

        if xk > yk:
            yield y
            try:
                y = ys.next()
                yk = y[0]
            except StopIteration:
                moreYs = False
            continue

        assert xk == yk
        yield (xk, x[1] + y[1])

        try:
            x = xs.next()
            xk = x[0]
        except StopIteration:
            moreXs = False
        try:
            y = ys.next()
            yk = y[0]
        except StopIteration:
            moreYs = False

    while moreXs:
        yield x
        try:
            x = xs.next()
        except StopIteration:
            moreXs = False

    while moreYs:
        yield y
        try:
            y = ys.next()
        except StopIteration:
            moreYs = False

def hist(zs, h, acgt):
    for z in zs:
        h[z[1]] = 1 + h.get(z[1], 0)
        acgt[z[0]&3] += 1
        yield z

def main(argv):
    opts = docopt.docopt(__doc__, argv)

    K = None

    out = opts['<output>']

    px = list(pairs(opts['<input>']))
    if len(px) == 1:
        with container(out, 'w') as z:
            h = {}
            acgt = [0, 0, 0, 0]
            ix = px[0]
            if len(ix) == 1:
                with container(ix[0], 'r') as z0:
                    K = z0.meta['K']
                    xs = readKmersAndCounts(z0)
                    zs = hist(xs, h, acgt)
                    writeKmersAndCounts(K, xs, z)
            else:
                with container(ix[0], 'r') as z0:
                    K = z0.meta['K']
                    xs = readKmersAndCounts(z0)
                    with container(ix[1], 'r') as z1:
                        K1 = z1.meta['K']
                        if K1 != K:
                            print >> sys.stderr, "mismatched K"
                            sys.exit(1)
                        ys = readKmersAndCounts(z1)
                        zs = hist(merge(xs, ys), h, acgt)
                        writeKmersAndCounts(K, zs, z)
            n = float(sum(acgt))
            acgt = [c/n for c in acgt]
            z.meta['hist'] = h
            z.meta['acgt'] = acgt
        return

    tmps = []
    tmpnm = tmpfile('.pmc')
    with container(tmpnm, 'w') as z:
        for ix in px:
            if len(ix) == 1:
                nm = 'tmp-' + str(len(tmps))
                tmps.append(nm)
                with container(ix[0], 'r') as z0:
                    if K is None:
                        K = z0.meta['K']
                    else:
                        K0 = z0.meta['K']
                        if K0 != K:
                            print >> sys.stderr, "mismatched K"
                            sys.exit(1)
                    xs = readKmersAndCounts(z0)
                    writeKmersAndCounts(K, xs, z, nm)
            else:
                nm = 'tmp-' + str(len(tmps))
                tmps.append(nm)
                with container(ix[0], 'r') as z0:
                    if K is None:
                        K = z0.meta['K']
                    else:
                        K0 = z0.meta['K']
                        if K0 != K:
                            print >> sys.stderr, "mismatched K"
                            sys.exit(1)
                    xs = readKmersAndCounts(z0)
                    with container(ix[1], 'r') as z1:
                        K1 = z1.meta['K']
                        if K1 != K:
                            print >> sys.stderr, "mismatched K"
                            sys.exit(1)
                        ys = readKmersAndCounts(z1)
                        writeKmersAndCounts(K, merge(xs, ys), z, nm)

    assert K is not None

    with container(out, 'w') as z:
        h = {}
        acgt = [0, 0, 0, 0]
        with container(tmpnm, 'r') as z0:
            zs = None
            for fn in tmps:
                xs = readKmersAndCounts(z0, fn)
                if zs is None:
                    zs = xs
                else:
                    zs = merge(zs, xs)
            zs = hist(zs, h, acgt)
            writeKmersAndCounts(K, zs, z)
        n = float(sum(acgt))
        acgt = [c/n for c in acgt]
        z.meta['hist'] = h
        z.meta['acgt'] = acgt

    os.remove(tmpnm)

if __name__ == '__main__':
    main(sys.argv[1:])



"""
Usage:
    zot ksnp [(-H DIST|-L DIST)] <output> <ref>

Use the kSNP algorithm (and extensions) to find candidate SNPs in a set
of reference k-mers.

Options:
    -H MAX          Perform Hamming based matching allowing a maximum
                    of MAX SNPs in the right hand side of k-mers.
    -L MAX          Perform Levenshtein based matching allowing a
                    maximum of MAX SNPs in the right hand side of
                    k-mers.
"""

import array
import sys

import docopt

from pykmer.basics import ham, lev
from pykmer.bits import popcnt
from pykmer.container import container
from pykmer.misc import unionfind
from pykmer.sparse import sparse

def ksnp(K, xs):
    J = (K - 1) // 2
    S = 2*(J + 1)
    M = (1 << (2*J)) - 1

    itms =  {}
    pfx = 0
    for x in xs:
        y = x >> S
        z = x & M
        if y != pfx:
            for ys in itms.values():
                if len(ys) > 1:
                    yield ys
            itms = {}
            pfx = y
        if z not in itms:
            itms[z] = []
        itms[z].append(x)
    for ys in itms.values():
        if len(ys) > 1:
            yield ys

def hamming(K, d, xs):
    J = (K - 1) // 2
    S = 2*(J + 1)
    T = 2*J
    M = (1 << (2*J)) - 1

    def grp(itms):
        z = len(itms)
        u = unionfind()
        for i in xrange(z):
            for j in xrange(i + 1, z):
                d0 = ham(itms[i], itms[j])
                if d0 <= d:
                    u.union(i, j)
        idx = {}
        for i in xrange(z):
            j = u.find(i)
            if j not in idx:
                idx[j] = []
            idx[j].append(i)
        for ys in idx.itervalues():
            if len(ys) == 1:
                continue
            zs = [itms[y] for y in ys]
            m = 0
            for z in zs:
                m |= 1 << ((z >> T) & 3)
            if popcnt(m) == 1:
                continue
            yield zs

    itms =  []
    pfx = 0
    for x in xs:
        y = x >> S
        if y != pfx:
            for g in grp(itms):
                yield g
            itms = []
            pfx = y
        itms.append(x)
    for g in grp(itms):
        yield g

def levenshtein(K, d, xs):
    J = (K - 1) // 2
    S = 2*(J + 1)
    T = 2*J
    M = (1 << (2*J)) - 1

    def grp(itms):
        z = len(itms)
        u = unionfind()
        for i in xrange(z):
            for j in xrange(i + 1, z):
                d0 = lev(K, itms[i], itms[j])
                if d0 <= d:
                    u.union(i, j)
        idx = {}
        for i in xrange(z):
            j = u.find(i)
            if j not in idx:
                idx[j] = []
            idx[j].append(i)
        for ys in idx.itervalues():
            if len(ys) == 1:
                continue
            zs = [itms[y] for y in ys]
            m = 0
            for z in zs:
                m |= 1 << ((z >> T) & 3)
            if popcnt(m) == 1:
                continue
            yield zs

    itms =  []
    pfx = 0
    for x in xs:
        y = x >> S
        if y != pfx:
            for g in grp(itms):
                yield g
            itms = []
            pfx = y
        itms.append(x)
    for g in grp(itms):
        yield g

def main(argv):
    opts = docopt.docopt(__doc__, argv)

    with container(opts['<ref>'], 'r') as z:
        K = z.meta['K']
        xs = readKmers(z)

        if opts['-H'] is not None:
            d = int(opts['-H'])
            ref = hamming(K, d, xs)
        elif opts['-L'] is not None:
            d = int(opts['-L'])
            ref = levenshtein(K, d, xs)
        else:
            ref = ksnp(K, xs)

        xs = []
        for ys in ref:
            xs += ys
    xs.sort()
    with container(opts['<output>'], 'w') as z:
        writeKmers(K, xs, z)

if __name__ == '__main__':
    main(sys.argv[1:])



"""
Usage:
    zot kmerize [-m MEM] [-D FRAC [-S SEED]] <k> <output> <input>...

Kmerize FASTA or FASTQ inputs to produce a standard container object.

Arguments:
    <k>         the length of the k-mers. Recommended values: 10-30
    <output>    the name of the output file.
                recommended naming convention
                    - mykmers.k25 for a k-mer set of 25-mers
                    - mykmers.kf25 for a k-mer frequency set of 25-mers
                    - mykmers.e25 for an expanded k-mer set of 25-mers

Options:
    -m MEM      in-memory buffer size (in MB)
    -D FRAC     subsample k-mers, using FRAC proportion of k-mers
    -S SEED     if -D is given, give a seed for determining the
                subspace (defaults to 0).
"""

import array
import gzip
import os
import subprocess
import sys
import time

import docopt

from pykmer.basics import kmersList, render, sub
from pykmer.container import container
from pykmer.container.std import readKmersAndCounts, writeKmersAndCounts
from pykmer.file import openFile, readFasta, readFastqBlock, tmpfile

from zotmer.library.merge import merge2

class KmerAccumulator:
    def __init__(self):
        self.toc = {}
        self.z = 0

    def __len__(self):
        return self.z

    def clear(self):
        self.toc = {}
        self.z = 0

    def add(self, x):
        xh = x >> 32
        xl = x & 0xFFFFFFFF
        if xh not in self.toc:
            self.toc[xh] = array.array('I')
        self.toc[xh].append(xl)
        self.z += 1

    def addList(self, xs):
        for x in xs:
            xh = x >> 32
            xl = x & 0xFFFFFFFF
            if xh not in self.toc:
                self.toc[xh] = array.array('I')
            self.toc[xh].append(xl)
            self.z += 1

    def kmers(self):
        xhs = self.toc.keys()
        xhs.sort()
        for xh in xhs:
            x0 = xh << 32
            xls = self.toc[xh].tolist()
            xls.sort()
            for xl in xls:
                x = x0 | xl
                yield x

    def stat(self):
        h = {}
        for (xh, xls) in self.toc.iteritems():
            l = len(xls)
            h[l] = 1 + h.get(l, 0)

        h = h.items()
        h.sort()
        for (l,c) in h:
            print '%d\t%d' % (l, c)

def pairs(xs):
    i = 0
    res = []
    while i + 1 < len(xs):
        res.append((xs[i], xs[i + 1]))
        i += 2
    if i < len(xs):
        res.append((xs[i], ))
    return res

def stripCompressionSuffix(nm):
    if nm.endswith('.gz'):
        return nm[:-3]
    return nm

def isFasta(nm):
    bnm = stripCompressionSuffix(nm)
    if bnm.endswith(".fa"):
        return True
    if bnm.endswith(".fasta"):
        return True
    if bnm.endswith(".fas"):
        return True
    if bnm.endswith(".fna"):
        return True

def mkParser(fn):
    if isFasta(fn):
        for (nm, seq) in readFasta(openFile(fn)):
            yield [(nm, seq)]
    else:
        for grps in readFastqBlock(openFile(fn)):
            yield [(grp[0], grp[1]) for grp in grps]

def mkPairs(xs):
    m = 0
    p = 0
    n = 0
    for x in xs:
        if x != p:
            if n > 0:
                m += 1
                yield (p, n)
            p = x
            n = 0
        n += 1
    if n > 0:
        m += 1
        yield (p, n)

def hist(zs, h):
    for z in zs:
        h[z[1]] = 1 + h.get(z[1], 0)
        yield z

def main(argv):
    opts = docopt.docopt(__doc__, argv)

    K = int(opts['<k>'])
    out = opts['<output>']
    Z = 1024*1024*32
    if opts['-m'] is not None:
        Z = 1024*1024*int(opts['-m'])

    buf = KmerAccumulator()
    n = 0
    tmps = []
    acgt = [0, 0, 0, 0]
    m = 0

    d = None
    if opts['-D'] is not None:
        d = float(opts['-D'])

        S = 0
        if opts['-S'] is not None:
            S = int(opts['-S'])

        cacheYes = set([])
        cacheNo = set([])

    tmpnm = tmpfile('.pmc')
    with container(tmpnm, 'w') as z:
        pass

    PN = 1024*1024

    nr = 0
    t0 = time.time()
    for fn in opts['<input>']:
        for rds in mkParser(fn):
            for (nm, seq) in rds:
                nr += 1
                if nr & (PN - 1) == 0:
                    t1 = time.time()
                    print >> sys.stderr, 'reads processed:', nr, (PN)/(t1 - t0), 'reads/second'
                    t0 = t1
                    #buf.stat()
                xs = kmersList(K, seq, True)
                if d is None:
                    buf.addList(xs)
                    for x in xs:
                        acgt[x&3] += 1
                        m += 1
                        n += 1
                else:
                    for x in xs:
                        if x in cacheNo:
                            continue
                        if x not in cacheYes:
                            if not sub(S, d, x):
                                cacheNo.add(x)
                                continue
                            cacheYes.add(x)
                        buf.add(x)
                        acgt[x&3] += 1
                        m += 1
                        n += 1
                    if len(cacheYes) > 1000000:
                        cacheYes = set([])
                    if len(cacheNo) > 1000000:
                        cacheNo = set([])
                if 8*n >= Z:
                    fn = 'tmps-%d' % (len(tmps),)
                    #print >> sys.stderr, "writing " + fn + "\t" + tmpnm
                    tmps.append(fn)
                    with container(tmpnm, 'a') as z:
                        writeKmersAndCounts(K, mkPairs(buf.kmers()), z, fn)
                    buf.clear()
                    n = 0

    t1 = time.time()
    print >> sys.stderr, 'reads processed:', nr, (nr % PN)/(t1 - t0), 'reads/second'

    if len(tmps) and len(buf):
        fn = 'tmps-%d' % (len(tmps),)
        #print >> sys.stderr, "writing " + fn + "\t" + tmpnm
        tmps.append(fn)
        with container(tmpnm, 'a') as z:
            writeKmersAndCounts(K, mkPairs(buf.kmers()), z, fn)
        buf = []

    while len(tmps) > 2:
        tmpnm2 = tmpfile('.pmc')
        tmps2 = []
        with container(tmpnm, 'r') as z0, container(tmpnm2, 'w') as z:
            ps = pairs(tmps)
            for p in ps:
                fn = 'tmps-%d' % (len(tmps2),)
                tmps2.append(fn)
                if len(p) == 1:
                    writeKmersAndCounts(K, readKmersAndCounts(z0, p[0]), z, fn)
                    continue
                h = {}
                merge2(z, K, readKmersAndCounts(z0, p[0]), readKmersAndCounts(z0, p[1]), h, fn)
        os.remove(tmpnm)
        tmpnm = tmpnm2
        tmps = tmps2

    with container(out, 'w') as z:
        h = {}
        if len(tmps) == 0:
            zs = hist(mkPairs(buf.kmers()), h)
            writeKmersAndCounts(K, zs, z)
        elif len(tmps) == 1:
            with container(tmpnm, 'r') as z0:
                writeKmersAndCounts(K, hist(readKmersAndCounts(z0, tmps[0]), h), z)
        else:
            assert len(tmps) == 2
            with container(tmpnm, 'r') as z0:
                merge2(z, K, readKmersAndCounts(z0, tmps[0]), readKmersAndCounts(z0, tmps[1]), h)
        n = float(sum(acgt))
        acgt = [c/n for c in acgt]
        z.meta['hist'] = h
        z.meta['acgt'] = acgt
        z.meta['reads'] = nr
    #os.remove(tmpnm)

if __name__ == '__main__':
    main(sys.argv[1:])



"""
Usage:
    zot mlst [-XK K] <alleles> <input>...

Determine which, out of a FASTA database of alleles exist in the
input set of k-mers.

Options:
    -X          Create an index
    -K K        If creating an index, use this value of K (default 27).
"""

import array
import cPickle
import sys

import docopt

from pykmer.container import container
from pykmer.container.std import readKmers
from pykmer.index import buildIndex, index

def main(argv):
    opts = docopt.docopt(__doc__, argv)

    if opts['-X']:
        K = 27
        if opts['-K']:
            K = int(opts['-K'])

        buildIndex(K, opts['<input>'], opts['<alleles>'])

        return

    idx = index(opts['<alleles>'])

    for inp in opts['<input>']:
        with container(inp, 'r') as z:
            K0 = z.meta['K']
            if K0 != idx.K:
                print >> sys.stderr, 'Input "%d" has different k to index' % (inp,)
                sys.exit(1)
            xs = readKmers(z)

            cs = [idx.lens[i] for i in xrange(len(idx.lens))]

            for x in xs:
                for j in idx[x]:
                    cs[j] -= 1

            for j in xrange(len(idx.lens)):
                assert cs[j] >= 0
                if cs[j] == 0:
                    print '%s\t%d\t%s' % (inp, j, idx.names[j])

if __name__ == '__main__':
    main(sys.argv[1:])



"""
Usage:
    zot vars [-r ref] <input>...

Options:
    -r ref      reference k-mers
"""

from pykmer.basics import can, fasta, lcp, render
from pykmer.file import readFasta
import pykmer.kset as kset
import pykmer.kfset as kfset
from pykmer.misc import heap
from pykmer.stats import logBinGe, logBinLe
from pykmer.exceptions import MismatchedK

import docopt
import math
import sys

def getK(ins):
    k = None
    for fn in ins:
        k0 = kfset.probeK(fn)
        if k is None:
            k = k0
        elif k != k0:
            raise MismatchedK(k, k0)
    return k

def group(K, J, n, xs):
    py = 0
    grp = []
    S = 2*(K-J)
    for (x,c) in xs:
        y = x >> S
        if y != py:
            if len(grp) > 0:
                yield (py, n, grp)
            py = y
            grp = []
        grp.append((x,c))
    if len(grp) > 0:
        yield (py, n, grp)

class Group:
    def __init__(self, K, J, n, xs):
        self.grps = group(K, J, n, xs)
        self.curr = None
        self.more = True
        self.next()

    def valid(self):
        return self.more

    def next(self):
        try:
            self.curr = self.grps.next()
        except StopIteration:
            self.more = False

    def this(self):
        if not self.valid():
            print self
        assert self.valid()
        return self.curr

    def __lt__(self, other):
        assert self.valid()
        assert other.valid()
        return self.this()[0] < other.this()[0]

    def __str__(self):
        return repr(self.curr) + '\t' + repr(self.more)

def main(argv):
    opts = docopt.docopt(__doc__, argv)

    K = getK(opts['<input>'])
    J = K - 1
    M = (1 << (2 * (K - J))) - 1

    if opts['-r'] is not None:
        refFn = opts['-r']

        (_, refXs) = kfset.read(refFn)
        xs = list(group(K, J, 0, refXs))

        for fn in opts['<input>']:
            (_, samXs) = kfset.read(fn)
            i = 0
            for (yCtx, _, yGrp) in group(K, J, 0, samXs):
                while i < len(xs) and xs[i][0] < yCtx:
                    i += 1
                assert i < len(xs)
                assert xs[i][0] == yCtx
                gt = float(sum([c for (x,c) in xs[i][2]]))
                gx = [0 for j in xrange(M+1)]
                for (x,c) in xs[i][2]:
                    gx[x&M] = c
                st = sum([c for (x,c) in yGrp])
                sx = [0 for j in xrange(M+1)]
                for (x,c) in yGrp:
                    sx[x&M] = c
                ss = []
                b = 0
                for j in xrange(M+1):
                    p = float(gx[j])/gt
                    v = 0.0
                    if 0.0 < p and p < 1.0:
                        v = logBinGe(p, st, sx[j])
                        if v < -10:
                            b |= 1 << j
                    ss.append('%3.2g' % (v,))
                if b > 0:
                    print '%s\t%s\t%s' % (render(J, yCtx), fasta(b), '\t'.join(ss))
                i += 1

        return

    # Parse files in parallel to get global distribution

    N = len(opts['<input>'])
    h = heap.heap()
    i = 0
    for fn in opts['<input>']:
        (_, xs) = kfset.read(fn)
        i += 1
        h.push(Group(K, J, i, xs))

    while len(h) > 0:
        xfs = []
        g = h.pop()
        gy = g.this()[0]
        xfs.append(g.this())
        g.next()
        if g.valid():
            h.push(g)
        for x in h.xs:
            assert x.valid()
        while len(h) > 0 and h.front().this()[0] == gy:
            g = h.pop()
            xfs.append(g.this())
            g.next()
            if g.valid():
                h.push(g)
            for i in xrange(len(h.xs)):
                assert h.xs[i].valid()

        ds = []
        gc = [0 for i in xrange(M+1)]
        for (_, n, xc) in xfs:
            t = sum([c for (x,c) in xc])
            d = [0 for i in xrange(M+1)]
            for (x,c) in xc:
                j = x & M
                gc[j] += c
                d[j] = c
            ds.append((n, d))

        res = ['*' for i in xrange(N)]
        seen = set([])
        gt = float(sum(gc))
        for (n, d) in ds:
            t = sum(d)
            b = [0 for i in xrange((M+1)/4)]
            for i in xrange(M+1):
                p = float(gc[i])/gt
                if 0.0 < p and p < 1.0:
                    #vL = logBinLe(p, t, d[i])
                    #vG = logBinGe(p, t, d[i])
                    #v = min(vL, vG)
                    v = logBinGe(p, t, d[i])
                    if v > -10:
                        w = i >> 2
                        j = i & 3
                        b[w] |= 1 << j
            res[n-1] = ''.join([fasta(b0) for b0 in b])
            seen.add(res[n-1])
        if len(seen) > 1:
            print '%s\t%s' % (render(J, gy), '\t'.join(res))

if __name__ == '__main__':
    main(sys.argv[1:])



"""
Usage:
    zot trim [-sc CUTOFF] <output> <input>

Options:
    -c CUTOFF   discard k-mers with frequency less than CUTOFF. A
                cutoff of 0 (the default) indicates that cutoff
                inference should be used.
    -s          force output to be a k-mer set
"""

import math
import sys

import docopt

from pykmer.container import container
from pykmer.container.std import readKmersAndCounts, writeKmersAndCounts

def sqr(x):
    return x * x

def gauss(x0, x, n):
    return math.exp(-sqr(x0 - x)/(2.0*n*n))

def smooth(h):
    h1 = {}
    for x0 in h.keys():
        num = 0
        den = 0
        for (xi,yi) in h.items():
            z = gauss(x0, xi, 1.5)
            num += z*yi
            den += z
        y0 = num/den
        h1[x0] = y0
    return h1

def infer(K, h):
    h1 = smooth(h)
    h1 = h1.items()
    h1.sort()
    v = h1[0]
    for v1 in h1[1:]:
        if v1[1] < v[1]:
            v = v1
        else:
            break
    return v[0]

def trim(xs, c):
    for (x,f) in xs:
        if f >= c:
            yield (x, f)

def main(argv):
    opts = docopt.docopt(__doc__, argv)

    inp = opts['<input>']
    out = opts['<output>']

    c = 0
    if opts['-c'] is not None:
        c = int(opts['-c'])

    with container(inp, 'r') as z:
        K = z.meta['K']
        h = z.meta['hist']
        if c == 0:
            c = infer(K, h)
            print >> sys.stderr, 'inferred cutoff:', c
        xs = readKmersAndCounts(z)
        with container(out, 'w') as w:
            w.meta = z.meta.copy()
            del w.meta['kmers']
            del w.meta['counts']
            writeKmersAndCounts(K, trim(xs, c), w)

if __name__ == '__main__':
    main(sys.argv[1:])



"""
Usage:
    zot dist [-M measure]... <k> <input>...

Options:
    -M measure  use "measure" for the distance between k-mer frequency sets.
                Use "-M list" to get a list of available measures.
"""

import pykmer.kfset as kfset
import pykmer.dist as dist
from pykmer.exceptions import MismatchedK

import array
import docopt
import fnmatch
import sys

measures = {}

class Measure:
    def __init__(self, name, desc, vec, func):
        self.name = name
        self.desc = desc
        self.vec = vec
        self.func = func

    def prep(self, K, fn):
        (meta, xs) = kfset.read(fn)
        fK = meta['K']
        if fK < K:
            raise MismatchedK(K, fK)
        if self.vec:
            S = 2*(fK - K)
            v = array.array('I', [0 for i in xrange(1 << (2*K))])
            for (x, c) in xs:
                y = x >> S
                v[y] += c
            return v
        
        S = 2*(fK - K)
        v = array.array('L', [])
        for (x,_) in xs:
            y = x >> S
            if len(v) == 0 or v[-1] != y:
                v.append(y)
        return v

    def measure(self, lhs, rhs):
        return self.func(lhs, rhs, self.vec)


def addMeasure(name, desc, vec, func):
    m = Measure(name, desc, vec, func)
    measures[name] = m

addMeasure('bray.curtis.quant',
           'Quantative Bray.Curtis distance', True, dist.brayCurtis)
addMeasure('bray.curtis.qual',
           'Qualitative Bray.Curtis distance', False, dist.brayCurtis)
addMeasure('chord.quant',
           'Quantative Chord distance', True, dist.chord)
addMeasure('chord.qual',
           'Qualitative Chord distance', False, dist.chord)
addMeasure('hellinger.quant',
           'Quantative Hellinger distance', True, dist.hellinger)
addMeasure('hellinger.qual',
           'Qualitative Hellinger distance', False, dist.hellinger)
addMeasure('jaccard.ab',
           'Abundance.based Jaccard distance', True, dist.jaccard)
addMeasure('jaccard.qual',
           'Qualitative Jaccard distance', False, dist.jaccard)
addMeasure('jensen.shannon',
           'Jensen.Shannon distance', True, dist.jensenShannon)
addMeasure('kulczynski.quant',
           'Quantative Kulczynski distance', True, dist.kulczynski)
addMeasure('kulczynski.qual',
           'Qualitative Kulczynski distance', False, dist.kulczynski)
addMeasure('ochiai.ab',
           'Abundance.based Ochiai distance', True, dist.ochiai)
addMeasure('ochiai.qual',
           'Qualitative Ochiai distance', False, dist.ochiai)
addMeasure('sorensen.ab',
           'Abundance.based Sorensen distance', True, dist.sorensen)
addMeasure('sorensen.qual',
           'Qualitative Sorensen distance', False, dist.sorensen)
addMeasure('whittaker.quant',
           'Quantative Whittaker distance', True, dist.whittaker)
addMeasure('whittaker.qual',
           'Qualitative Whittaker distance', False, dist.whittaker)

def main(argv):
    opts = docopt.docopt(__doc__, argv)

    if "list" in opts['-M']:
        ms = measures.keys()
        ms.sort()
        msg = []
        for m in ms:
            msg.append(m + '\t' + measures[m].desc)
        print '\n'.join(msg)
        return

    allMs = measures.keys()
    allMs.sort()

    seen = set([])
    bad = False
    for mo in opts['-M']:
        found = False
        for m in allMs:
            if fnmatch.fnmatch(m, mo):
                seen.add(m)
                found = True
        if not found:
            print >> sys.stderr, 'warning: measure \'%s\' not found. Use -M list to see all measures.' % (mo,)
            bad = True
    ms = list(seen)
    ms.sort()

    if len(ms) == 0 or bad:
        return

    K = int(opts['<k>'])

    N = len(opts['<input>'])
    fns = opts['<input>']

    hdr = ['lhs.name', 'rhs.name']
    fmt = ['%s', '%s']
    vecNeeded = None
    setNeeded = None
    for m in ms:
        hdr.append(m)
        fmt.append('%g')
        if measures[m].vec:
            vecNeeded = m
        else:
            setNeeded = m
    fmt = '\t'.join(fmt)

    print '\t'.join(hdr)
    for i in xrange(N):
        lhsVec = None
        if vecNeeded is not None:
            lhsVec = measures[vecNeeded].prep(K, fns[i])
        lhsSet = None
        if setNeeded is not None:
            lhsSet = measures[setNeeded].prep(K, fns[i])

        for j in xrange(i + 1, N):
            rhsVec = None
            if vecNeeded is not None:
                rhsVec = measures[vecNeeded].prep(K, fns[j])
            rhsSet = None
            if setNeeded is not None:
                rhsSet = measures[setNeeded].prep(K, fns[j])

            vs = [fns[i], fns[j]]
            for m in ms:
                if measures[m].vec:
                    d = measures[m].measure(lhsVec, rhsVec)
                else:
                    d = measures[m].measure(lhsSet, rhsSet)
                vs.append(d)
            print fmt % tuple(vs)

if __name__ == '__main__':
    main(sys.argv[1:])



"""
Usage:
    zot scan [-AX] <genes> <input>...

Options:
    -A          print all alleles
    -X          create an index
"""

from pykmer.basics import fasta, kmersWithPos, ham, lcp, rc, render
from pykmer.container import container
from pykmer.container.vectors import read64, write64, read32, write32, read32s, write32s
from pykmer.container.std import readKmers
from pykmer.file import readFasta
from pykmer.misc import unionfind
from pykmer.sparse import sparse
from pykmer.stats import counts2cdf, ksDistance2, log1mexp, logAdd, logChoose, logFac

import array
import cPickle
import docopt
import math
import sys

def lev(x, y):
    zx = len(x)
    zy = len(y)
    buf0 = array.array('I', [0 for i in xrange(zy + 1)])
    buf1 = array.array('I', [0 for i in xrange(zy + 1)])

    v0 = buf0
    v1 = buf1

    for i in xrange(zy):
        v0[i] = i

    for i in xrange(zx):
        v1[0] = i + 1
        for j in xrange(zy):
            if x[i] == y[j]:
                c = 0
            else:
                c = 1
            a1 = v0[j] + c
            a2 = v1[j] + 1
            a3 = v0[j + 1] + 1
            v1[j + 1] = min([a1, a2, a3])
        t = v1
        v1 = v0
        v0 = t
    return v0[zy]

def resolveLcp(K, S, L, Y, x):
    z = S.rank(x)
    for k in xrange(z):
        j = z - k - 1
        l = lcp(K, x, S.select(j))
        if l > L[j]:
            L[j] = l
            Y[j] = x
        else:
            break
    for j in xrange(z, S.count()):
        l = lcp(K, x, S.select(j))
        if l > L[j]:
            L[j] = l
            Y[j] = x
        else:
            break

def resolveAll(K, S, L, Y, xs):
    i = 0
    Z = S.count()
    for x in xs:
        while i < Z and S.select(i) <= x:
            y = S.select(i)
            l = lcp(K, x, y)
            if l > L[i]:
                L[i] = l
                Y[i] = x
            i += 1
    while i < Z:
        y = S.select(i)
        l = lcp(K, x, y)
        if l > L[i]:
            L[i] = l
            Y[i] = x
        i += 1

    for i in xrange(1, Z):
        x = Y[i-1]
        y = S.select(i)
        l = lcp(K, x, y)
        if l > L[i]:
            L[i] = l
            Y[i] = x

def succ(K, xs, x):
    m = (1 << (2*K)) - 1
    y0 = (x << 2) & m
    y1 = y0 + 4
    (r0, r1) = xs.rank2(y0, y1)
    return range(r0, r1)

def null(g, s, j):
    if j == 0:
        return 0.0
    return math.exp(s*math.log1p(-math.pow(g,j)))

def logNull(g, s, j):
    return s*math.log1p(-math.pow(g,j))

def logGammaPInt(n, x):
    if x < n:
        lx = math.log(x)
        w = 0
        v = logFac(n)
        s = w - v
        j = 1
        while True:
            w += lx
            v += math.log(j+n)
            u = logAdd(s, w - v)
            if s == u:
                break
            s = u
            j += 1
        return n*lx + s - x
    else:
        lx = math.log(x)
        s = 0
        v = 0
        for k in xrange(1, n):
            v += math.log(k)
            t = k*lx - v
            u = logAdd(t, s)
            s = u
        w = min(s - x, -1e-200)
        return log1mexp(w)

def chi2(n, x):
    d = int(math.ceil(n/2.0))
    c2 = x/2.0
    return logGammaPInt(d, c2)

def uniq(xs):
    px = None
    for x in xs:
        if x != px:
            yield x
            px = x

def main(argv):
    opts = docopt.docopt(__doc__, argv)

    if opts['-X']:
        K = 27
        S = []
        N = 0
        qacgt = [0, 0, 0, 0]
        for fn in opts['<input>']:
            with open(fn) as f:
                for (nm, seq) in readFasta(f):
                    if len(seq) < K:
                        continue
                    for (x,p) in kmersWithPos(K, seq, True):
                        S.append(x) 
                        qacgt[x&3] += 1
                        N += 1
        S.sort()
        qacgt = [float(c)/float(N) for c in qacgt]
        S = sparse(2*K, array.array('L', uniq(S)))
        lens = array.array('I', [])
        nms = []
        seqs = []
        n = 0
        tmp = [[] for i in xrange(S.count())]
        for fn in opts['<input>']:
            with open(fn) as f:
                for (nm, seq) in readFasta(f):
                    if len(seq) < K:
                        print >> sys.stderr, "warning: `%s' skipped" % (nm,)
                        continue
                    nms.append(nm)
                    seqs.append(seq)
                    lens.append(len(seq))
                    for (x,p) in kmersWithPos(K, seq, True):
                        r = S.rank(x)
                        tmp[r].append((n, p))
                    n += 1
        T = array.array('I', [])
        U = array.array('I', [])
        V = array.array('i', [])
        t = 0
        for nps in tmp:
            T.append(t)
            t += len(nps)
            for (n, p) in nps:
                U.append(n)
                V.append(p)
        T.append(t)
        del tmp

        gfn = opts['<genes>']
        with container(gfn, 'w') as z:
            z.meta['K'] = K
            z.meta['S'] = S.count()
            write64(z, S.xs, 'S')
            z.meta['T'] = len(T)
            write64(z, T, 'T')
            z.meta['U'] = len(U)
            write32(z, U, 'U')
            z.meta['V'] = len(V)
            write32s(z, V, 'V')
            z.meta['lens'] = lens
            z.meta['qacgt'] = qacgt
            z.meta['nms'] = nms
            z.meta['seqs'] = seqs

        return

    print >> sys.stderr, "loading..."

    gfn = opts['<genes>']
    with container(gfn, 'r') as z:
        K = z.meta['K']
        S = array.array('L', read64(z, 'S', z.meta['S']))
        S = sparse(2*K, S)
        T = array.array('L', read64(z, 'T', z.meta['T']))
        U = array.array('I', read32(z, 'U', z.meta['U']))
        V = array.array('i', read32s(z, 'V', z.meta['V']))
        lens = z.meta['lens']
        qacgt = z.meta['qacgt']
        nms = z.meta['nms']
        seqs = z.meta['seqs']

    print >> sys.stderr, "done."

    for fn in opts['<input>']:
        L = array.array('B', [0 for i in xrange(S.count())])
        Y = array.array('L', [0 for i in xrange(S.count())])
        with container(fn, 'r') as z:
            sacgt = z.meta['acgt']
            xs = readKmers(z)
            X = array.array('L', xs)
        M = len(X)
        resolveAll(K, S, L, Y, X)
        X = sparse(2*K, X)

        g = sum([qp*sp for (qp, sp) in zip(qacgt, sacgt)])
        print >> sys.stderr, "g =", g
        nm = [null(g, M, j) for j in range(0, K+1)]

        # counts for computing distribution of prefix lengths
        cnt = [[0 for j in xrange(K+1)] for i in xrange(len(nms))]

        # the k-mers that we pulled by lcp from the sample
        # for each position of each query.
        P = [array.array('L', [0 for j in xrange(lens[i] - K + 1)]) for i in xrange(len(lens))]

        # the length of the lcp for each position of each query.
        Q = [array.array('B', [0 for j in xrange(lens[i] - K + 1)]) for i in xrange(len(lens))]

        for i in xrange(S.count()):
            for j in xrange(T[i], T[i+1]):
                n = U[j]
                p = V[j]
                y = Y[i]
                l = L[i]
                cnt[n][l] += 1
                if p > 0:
                    p -= 1
                else:
                    p = -(p + 1)
                    y = rc(K, y)
                if l > Q[n][p]:
                    Q[n][p] = l
                    P[n][p] = y

        for i in xrange(len(nms)):
            # iterate over the queries

            qc = math.log(K*0.05/float(lens[i] - K + 1)/2)

            # Link up "de Bruijn" sequences
            m = (1 << (2*K - 2)) - 1
            py = 0
            u = unionfind()
            for j in xrange(lens[i] - K + 1):
                x = P[i][j]
                y = x >> 2
                if j > 0:
                    d = ham(py, y)
                    if d == 0:
                        u.union(j-1, j)
                py = x & m

            # Gather up the de Bruin fragments
            udx = {}
            for j in xrange(lens[i] - K + 1):
                v = u.find(j)
                if v not in udx:
                    udx[v] = []
                udx[v].append(j)

            # Index the left hand k-mers
            idxLhs = {}
            kx = []
            for (jx, js) in udx.iteritems():
                q = 0
                for j in js:
                    q += math.log1p(-nm[Q[i][j]])
                if q > math.log(0.05/len(js)):
                    continue
                kx.append((-len(js), jx))
                idxLhs[P[i][js[0]]] = jx
            kx.sort()

            # Attempt to link up fragments
            links = {}
            for (_, jx) in kx:
                jR = udx[jx][-1]
                if jR == lens[i] - K + 1:
                    continue
                x = P[i][jR]
                xs = []
                lnk = None
                for k in xrange(100):
                    ys = succ(K, X, x)
                    if len(ys) != 1:
                        break
                    x = ys[0]
                    if x in idxLhs:
                        lnk = idxLhs[x]
                        break
                    xs.append(x)
                if lnk is not None:
                    links[jx] = xs
                    u.union(jx, lnk)

            # Gather up the linked fragments
            vdx = {}
            for j in [jx for (_, jx) in kx]:
                v = u.find(j)
                if v not in vdx:
                    vdx[v] = []
                vdx[v].append(j)

            res = []
            for (jxx, jxs) in vdx.iteritems():
                # Order the gragments by start position
                fs = [(udx[jx][0], jx) for jx in jxs]
                fs.sort()
                sxs = []
                for fj in xrange(len(fs)):
                    (_, jx) = fs[fj]
                    beg = udx[jx][0]
                    end = udx[jx][-1] + 1
                    if fj == 0:
                        for j in xrange(beg):
                            sxs.append((0, 0))
                    xs = links.get(jx, None)
                    for j in xrange(beg, end):
                        x = P[i][j]
                        l = Q[i][j]
                        sxs.append((x, l))
                    if xs:
                        for x in xs:
                            sxs.append((x, 27))
                    else:
                        if fj < len(fs) - 1:
                            nxt = fs[fj+1][0]
                        else:
                            nxt = lens[i] - K + 1
                        for j in xrange(end, nxt):
                            sxs.append((0, 0))
                seq = [[0, 0, 0, 0] for j in xrange(len(sxs) + K - 1)]
                for j in xrange(len(sxs)):
                    (x, l) = sxs[j]
                    p = math.log1p(-nm[l])
                    for k in xrange(K):
                        seq[j + K - k - 1][x&3] += p
                        x >>= 2
                ax = []
                p = None
                inf = False
                for j in xrange(len(seq)):
                    b = 0
                    for k in xrange(4):
                        if seq[j][k] < qc:
                            b |= 1 << k
                    ax.append(fasta(b))
                    ssj = sum(seq[j])
                    if p is None:
                        p = ssj
                    else:
                        p = logAdd(p, ssj)
                    if ssj > -1e-300:
                        inf = True
                dst = counts2cdf(cnt[i])
                (_, kd) = ksDistance2(dst, nm)
                df = math.ceil(len(seq)/float(K))
                if inf:
                    q = 1e300
                    pv = 0.0
                else:
                    q = 2*math.exp(p)
                    pv = chi2(df, q)
                res.append((pv, q, kd, ''.join(ax)))

            if len(res) == 0:
                continue

            res.sort()
            if res[0][0] < -2:
                #ed = lev(seqs[i], res[0][2])
                ed = 0
                pv = res[0][0]/math.log(10)
                c2 = res[0][1]
                kd = res[0][2]
                a = res[0][3]
                print '%d\t%d\t%d\t%g\t%g\t%g\t%s\t%s' % (i, lens[i], len(a), kd, c2, pv, nms[i], a)
            sys.stdout.flush()

if __name__ == '__main__':
    main(sys.argv[1:])



"""
Usage:
    zot project <ref> <output> <input>

Project one or more inputs on to a reference set. For each k-mer in <ref>,
a whitespace separated 0 or a 1 is printed indicating whether that k-mer
was present in the input, with a separate line for each input k-mer set.
"""

import array
import sys

import docopt

from pykmer.container import container
from pykmer.container.std import readKmers, writeKmers, readKmersAndCounts, writeKmersAndCounts

def project1(xs, ys):
    i = 0
    Z = len(xs)
    for y in ys:
        while i < Z and xs[i] < y:
            i += 1
        if i == Z:
            break
        if xs[i] == y:
            yield y
            i += 1

def project2(xs, ys):
    i = 0
    Z = len(xs)
    for y in ys:
        while i < Z and xs[i] < y[0]:
            i += 1
        if i == Z:
            break
        if xs[i] == y[0]:
            yield y
            i += 1

def main(argv):
    opts = docopt.docopt(__doc__, argv)

    with container(opts['<ref>'], 'r') as z:
        K = z.meta['K']
        xs = array.array('L', readKmers(z))
    Z = len(xs)

    with container(opts['<input>'], 'r') as z0:
        K0 = z0.meta['K']
        if K0 != K:
            print >> sys.stderr, "mismatched K (%d)" % (K0, )
            sys.exit(1)

        with container(opts['<output>'], 'w') as z:
            if 'counts' in z0.meta:
                ys = readKmersAndCounts(z0)
                writeKmersAndCounts(K, project2(xs, ys), z)
            else:
                ys = readKmers(z0)
                writeKmers(K, project1(xs, ys), z)

if __name__ == '__main__':
    main(sys.argv[1:])



"""
Usage:
    zot dump <input>
"""

from pykmer.basics import render
from pykmer.container import container
from pykmer.container.std import readKmers, readKmersAndCounts

import docopt
import sys

def main(argv):
    opts = docopt.docopt(__doc__, argv)

    inp = opts['<input>']
    with container(inp, 'r') as z:
        K = z.meta['K']
        if 'kmers' not in z.meta:
            print >> sys.stderr, 'cannot dump "%s" as it contains no k-mers' % (inp,)
            return
        if 'counts' in z.meta:
            xs = readKmersAndCounts(z)
            for (x, c) in xs:
                print '%s\t%d' % (render(K, x), c)
        else:
            xs = readKmers(z)
            for x in xs:
                print render(K, x)

if __name__ == '__main__':
    main(sys.argv[1:])



import pykmer.container.std as std

def merge2(z, K, xs, ys, h, nm = None):
    if nm is not None:
        knm = nm + '-kmers'
        cnm = nm + '-counts'
    else:
        knm = None
        cnm = None

    with std.kmerWriter(z, K, knm) as kx, std.countsWriter(z, K, cnm) as cx:
        moreXs = True
        moreYs = True

        try:
            x = xs.next()
            xk = x[0]
        except StopIteration:
            moreXs = False
        try:
            y = ys.next()
            yk = y[0]
        except StopIteration:
            moreYs = False

        while moreXs and moreYs:
            if xk < yk:
                h[x[1]] = 1 + h.get(x[1], 0)
                kx.append(xk), cx.append(x[1])
                try:
                    x = xs.next()
                    xk = x[0]
                except StopIteration:
                    moreXs = False
                continue

            if xk > yk:
                h[y[1]] = 1 + h.get(y[1], 0)
                kx.append(yk), cx.append(y[1])
                try:
                    y = ys.next()
                    yk = y[0]
                except StopIteration:
                    moreYs = False
                continue

            assert xk == yk
            c = x[1] + y[1]
            h[c] = 1 + h.get(c, 0)
            kx.append(xk), cx.append(c)

            try:
                x = xs.next()
                xk = x[0]
            except StopIteration:
                moreXs = False
            try:
                y = ys.next()
                yk = y[0]
            except StopIteration:
                moreYs = False

        while moreXs:
            h[x[1]] = 1 + h.get(x[1], 0)
            kx.append(x[0]), cx.append(x[1])
            try:
                x = xs.next()
            except StopIteration:
                moreXs = False

        while moreYs:
            h[y[1]] = 1 + h.get(y[1], 0)
            kx.append(y[0]), cx.append(y[1])
            try:
                y = ys.next()
            except StopIteration:
                moreYs = False



"""
Usage:
    zot jaccard [-abp P] <input>...

Compute Jaccard indexes between k-mer sets. By default, indexes are
computed only between the first k-mer set and all the remaining
k-mer sets. If the -a option is given, all pairwise indexes are
computed.  If the -p P option is given, a Null hypothesis test is
performed for the hypothesis that the underlying Jaccard Index is
less than P. This is particularly useful if subsets of k-mers are
being used (NB, if the k-mer sets are large, the statistics can be
very expensive to compute).

Options:
    -a          print all pairwise distances
    -p P        Jaccard distance thresshhold for p-value computation
"""

import array
import math
import sys

import docopt

from pykmer.container import container
from pykmer.container.std import readKmers
from pykmer.stats import logAdd, logChoose

def jaccard(xs, ys):
    xz = len(xs)
    yz = len(ys)
    i = 0
    j = 0
    d = 0
    b = 0
    while i < xz and j < yz:
        x = xs[i]
        y = ys[j]
        if x < y:
            d += 1
            i += 1
            continue
        if x > y:
            d += 1
            j += 1
            continue
        b += 1
        i += 1
        j += 1
    d += xz - i
    d += yz - j
    return (b, b+d, float(b) / float(b + d))

def logIx(x, m, n):
    lx = math.log(x)
    j = m
    v = logChoose(n + j - 1, j)
    s = v + j * lx
    while True:
        j += 1
        v += math.log((n + j - 1.0) / j)
        t = v + j * lx
        #print j, t, s
        u = logAdd(s, t)
        if u == s:
            break
        s = u
    return n*math.log1p(-x) + s

def quantBeta(q, m, n):
    lq = math.log(q)
    l = 1e-10
    h = 1 - 1e-10
    while (h - l) > 1e-7:
        x = (h + l) / 2.0
        lp = logIx(x, m, n)
        if lp < lq:
            l = x
        else:
            h = x
    return l

def main(argv):
    opts = docopt.docopt(__doc__, argv)

    fns = opts['<input>']

    Z = 1
    if opts['-a']:
        Z = len(fns)

    p = None
    if opts['-p'] is not None:
        p = float(opts['-p'])

    for i in xrange(Z):
        with container(fns[i], 'r') as z0:
            xK = z0.meta['K']
            xs = array.array('L', readKmers(z0))
            for j in xrange(i + 1, len(fns)):
                with container(fns[j], 'r') as z1:
                    yK = z1.meta['K']
                    ys = array.array('L', readKmers(z1))
                    if xK != yK:
                        print >> sys.stderr, 'mismatched K:', fns[j]
                        sys.exit(1)
                    (isec, union, d) = jaccard(xs, ys)
                    if p is None:
                        print '%s\t%s\t%d\t%d\t%d\t%d\t%f' % (fns[i], fns[j], len(xs), len(ys), isec, union, d)
                    else:
                        pv = logIx(p, isec+1, (union - isec) + 1) / math.log(10)
                        q05 = quantBeta(0.05, isec+1, (union - isec) + 1)
                        q95 = quantBeta(0.95, isec+1, (union - isec) + 1)
                        print '%s\t%s\t%d\t%d\t%d\t%d\t%f\t-%f\t+%f\t%f' % (fns[i], fns[j], len(xs), len(ys), isec, union, d, d - q05, q95 - d, pv)
                    sys.stdout.flush()

if __name__ == '__main__':
    main(sys.argv[1:])



"""
Usage:
    zot info <input>...
"""

from pykmer.container import probe

import docopt

def main(argv):
    opts = docopt.docopt(__doc__, argv)

    for inp in opts['<input>']:
        (m, _) = probe(inp)
        print m

if __name__ == '__main__':
    main(sys.argv[1:])



"""
Usage:
    zot sample [-DS SEED] [-P PROBABILITY] <output> <input>

Options:
    -D              use deterministic sampling
    -P PROBABILITY  the proportion of samples to include in the output.
                    default: 0.01
    -S SEED         use the given seed for the sampling
"""

import random
import sys

import docopt

from pykmer.basics import murmer
from pykmer.container import container
from pykmer.container.std import readKmersAndCounts, writeKmersAndCounts

def sampleR(p, xs, h):
    for x in xs:
        if random.random() < p:
            h[x[1]] = 1 + h.get(x[1], 0)
            yield x

def sampleD(p, s, xs, h):
    M = 0xFFFFFFFFFF
    for x in xs:
        y = x[0]
        u = float(murmer(y, s)&M)/float(M)
        if u < p:
            h[x[1]] = 1 + h.get(x[1], 0)
            yield x

def main(argv):
    opts = docopt.docopt(__doc__, argv)

    p = 0.01
    if opts['-P'] is not None:
        p = float(opts['-P'])
    inp = opts['<input>']
    out = opts['<output>']
    with container(out, 'w') as z:
        h = {}
        with container(inp, 'r') as z0:
            K = z0.meta['K']
            z.meta = z0.meta.copy()
            del z.meta['kmers']
            del z.meta['counts']
            xs = readKmersAndCounts(z0)
            if opts['-D'] is None:
                if opts['-S'] is not None:
                    S = long(opts['-S'])
                    random.seed(S)
                    writeKmersAndCounts(K, sampleR(p, xs, h), z)
            else:
                S = 0
                if opts['-S'] is not None:
                    S = long(opts['-S'])
                    writeKmersAndCounts(K, sampleD(p, S, xs, h), z)
        z.meta['hist'] = h

if __name__ == '__main__':
    main(sys.argv[1:])



"""
Usage:
    zot spoligo [-lp PROBES] <input>...

Perform spoligotyping on sets of k-mers. By default, use the MTB
spoligotyping probes. If the -p option is used, then the argument
is a file name of a text file continaining an optional name and a
probe on each line. Lines starting with '#' are ignored.

Options:
    -l          use long output format with one line per probe
    -p PROBES   the name of a file containing probe sequences
"""

import array
import sys

import docopt

from pykmer.basics import kmer, render
from pykmer.container import container
from pykmer.container.std import readKmers
from pykmer.misc import uniq
from pykmer.sparse import sparse

def neigh(K, x, d):
    if d == 0:
        return []
    xs = []
    for i in xrange(K):
        for j in xrange(3):
            xs.append(x ^ ((j + 1) << (2*i)))
    xs.sort()

    if d == 1:
        return xs

    zs = []
    for y in xs:
        zs += neigh(K, y, d - 1)
    zs.sort()
    uniq(zs)
    diff(zs, [x])
    diff(zs, xs)
    return zs

def probe(s):
    return (len(s), kmer(s))

def findApprox(J, x, K, xs, D):
    assert J <= K

    s = 2*(K - J)
    y0 = x << s
    y1 = (x+1) << s
    (r0, r1) = xs.rank2(y0, y1)
    if r1 - r0 > 0:
        return True

    for d in xrange(1, D+1):
        for y in neigh(J, x, d):
            y0 = y << s
            y1 = (y+1) << s
            (r0, r1) = xs.rank2(y0, y1)
            if r1 - r0 > 0:
                return True
    return False

def findProbe(p, K, xs):
    Kp = p[0]
    x = p[1]
    if Kp <= K:
        return findApprox(Kp, x, K, xs, 2)

    # Kp > K.
    M = (1 << (2*K)) - 1
    y0 = x
    ys = []
    for i in xrange(1 + Kp - K):
        y = y0 & M
        if not findApprox(K, y, K, xs, 2):
            return False
        y0 >>= 2
    return True

probesMTB = [
    probe('ATAGAGGGTCGCCGGTTCTGGATCA'),
    probe('CCTCATAATTGGGCGACAGCTTTTG'),
    probe('CCGTGCTTCCAGTGATCGCCTTCTA'),
    probe('ACGTCATACGCCGACCAATCATCAG'),
    probe('TTTTCTGACCACTTGTGCGGGATTA'),
    probe('CGTCGTCATTTCCGGCTTCAATTTC'),
    probe('GAGGAGAGCGAGTACTCGGGGCTGC'),
    probe('CGTGAAACCGCCCCCAGCCTCGCCG'),
    probe('ACTCGGAATCCCATGTGCTGACAGC'),
    probe('TCGACACCCGCTCTAGTTGACTTCC'),
    probe('GTGAGCAACGGCGGCGGCAACCTGG'),
    probe('ATATCTGCTGCCCGCCCGGGGAGAT'),
    probe('GACCATCATTGCCATTCCCTCTCCC'),
    probe('GGTGTGATGCGGATGGTCGGCTCGG'),
    probe('CTTGAATAACGCGCAGTGAATTTCG'),
    probe('CGAGTTCCCGTCAGCGTCGTAAATC'),
    probe('GCGCCGGCCCGCGCGGATGACTCCG'),
    probe('CATGGACCCGGGCGAGCTGCAGATG'),
    probe('TAACTGGCTTGGCGCTGATCCTGGT'),
    probe('TTGACCTCGCCAGGAGAGAAGATCA'),
    probe('TCGATGTCGATGTCCCAATCGTCGA'),
    probe('ACCGCAGACGGCACGATTGAGACAA'),
    probe('AGCATCGCTGATGCGGTCCAGCTCG'),
    probe('CCGCCTGCTGGGTGAGACGTGCTCG'),
    probe('GATCAGCGACCACCGCACCCTGTCA'),
    probe('CTTCAGCACCACCATCATCCGGCGC'),
    probe('GGATTCGTGATCTCTTCCCGCGGAT'),
    probe('TGCCCCGGCGTTTAGCGATCACAAC'),
    probe('AAATACAGGCTCCACGACACGACCA'),
    probe('GGTTGCCCCGCGCCCTTTTCCAGCC'),
    probe('TCAGACAGGTTCGCGTCGATCAAGT'),
    probe('GACCAAATAGGTATCGGCGTGTTCA'),
    probe('GACATGACGGCGGTGCCGCACTTGA'),
    probe('AAGTCACCTCGCCCACACCGTCGAA'),
    probe('TCCGTACGCTCGAAACGCTTCCAAC'),
    probe('CGAAATCCAGCACCACATCCGCAGC'),
    probe('CGCGAACTCGTCCACAGTCCCCCTT'),
    probe('CGTGGATGGCGGATGCGTTGTGCGC'),
    probe('GACGATGGCCAGTAAATCGGCGTGG'),
    probe('CGCCATCTGTGCCTCATACAGGTCC'),
    probe('GGAGCTTTCCGGCTTCTATCAGGTA'),
    probe('ATGGTGGGACATGGACGAGCGCGAC'),
    probe('CGCAGAATCGCACCGGGTGCGGGAG')
]

def diff(xs, ys):
    zX = len(xs)
    zY = len(ys)
    i = 0
    j = 0
    k = 0
    while i < zX and j < zY:
        assert k <= i
        x = xs[i]
        y = ys[j]
        if x < y:
            if k != i:
                xs[k] = xs[i]
            i += 1
            k += 1
            continue
        if x > y:
            j += 1
            continue
        i += 1
        j += 1
    if k != i:
        while i < zX:
            xs[k] = xs[i]
            i += 1
            k += 1
        del xs[k:]

def main(argv):
    opts = docopt.docopt(__doc__, argv)

    nms = [str(i+1) for i in xrange(len(probesMTB))]
    probes = probesMTB

    if opts['-p'] is not None:
        nms = []
        probes = []
        bad = False
        with open(opts['-p']) as f:
            i = 0
            ln = 0
            for l in f:
                ln += 1
                if l[0] == '#':
                    continue
                i += 1
                t = l.split()
                if len(t) == 1:
                    nms.append(str(i))
                    probes.append(probe(t[0]))
                elif len(t) == 2:
                    nms.append(t[0])
                    probes.append(probe(t[1]))
                else:
                    bad = True
                    print >> sys.stderr, '%s line %d, badly formatted.' % (opts['-p'], i)
        if bad:
            sys.exit(1)

    for inp in opts['<input>']:
        with container(inp, 'r') as z:
            K = z.meta['K']
            xs = readKmers(z)
            xs = sparse(2*K, array.array('L', xs))

        res = []
        for i in xrange(len(probes)):
            if findProbe(probes[i], K, xs):
                res.append('1')
            else:
                res.append('0')
        if opts['-l']:
            for i in xrange(len(nms)):
                print '%s\t%s\t%s' % (inp, nms[i], res[i])
        else:
            print inp + '\t' + ''.join(res)

if __name__ == '__main__':
    main(sys.argv[1:])



from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand

class PyTest(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        #import here, cause outside the eggs aren't loaded
        import pytest
        pytest.main(self.test_args)

setup(name='zotmer',
        version='0.1',
        description='A python k-mer application workbench',
        url='http://github.com/drtconway/zotmer',
        author='Tom Conway',
        author_email='drtomc@gmail.com',
        license='Apache2',
        keywords='bioinformatics genomics pathogenomics',
        packages=find_packages(),
        install_requires=['docopt', 'pykmer>=0.1'],
        entry_points = {
            'console_scripts' : ['zot=zotmer.cli:main']
        },
        tests_require=['pytest'],
        cmdclass = {'test': PyTest},
        zip_safe=False)



"""
Usage:
    zot [options] <command> [<args>...]

options:
    --help          print usage information
    -V, --version   print version information
"""

import importlib
import os
import pkgutil
import sys

import docopt

from pykmer.file import autoremove

import commands

def mainInner():
    args = docopt.docopt(__doc__,
                         version='Zotmer k-mer toolkit 0.1',
                         options_first=True)

    if args['<command>'] == 'help' and len(args['<args>']) != 1:
        print __doc__
        print "Available commands:"
        for _, name, is_pkg in pkgutil.iter_modules([commands.__path__[0]]):
            print '\t' + name
        print '\nuse "zot help <command>" for command specific help.'
        return 0

    if args['<command>'] == 'help' and len(args['<args>']) == 1:
        modname = commands.__name__ + '.' + args['<args>'][0]
        try:
            argv = [args['<command>']] + args['<args>']
            m = importlib.import_module(modname)
            print m.__doc__
            return 0
        except ImportError:
            print >> sys.stderr, "unable to load command `%s', use `zot help` for help." % (args['<command>'], )
            return 1

    modname = commands.__name__ + '.' + args['<command>']
    try:
        argv = [args['<command>']] + args['<args>']
        m = importlib.import_module(modname)
        return m.main(argv)
    except ImportError:
        print >> sys.stderr, "unable to load command `%s', use `zot help` for help." % (args['<command>'], )
        return 1

def main():
    try:
        with autoremove():
            mainInner()
    except KeyboardInterrupt:
        pass

if __name__ == '__main__':
    main()




"""
Usage:
    zot hist <input>...

Options:
    -u              update the input container to include the histogram
"""

from pykmer.container import container

import docopt
import sys

def main(argv):
    opts = docopt.docopt(__doc__, argv)

    for inp in opts['<input>']:
        with container(inp, 'r') as z:
            if 'hist' in z.meta:
                h = z.meta['hist'].items()
                h.sort()
                for (f,c) in h:
                    print '%s\t%d\t%d' % (inp, f, c)

if __name__ == '__main__':
    main(sys.argv[1:])
